package cache

import (
	"context"
	"errors"
	"gitea.loveuer.com/yizhisec/packages/tool"
	"github.com/go-redis/redis/v8"
	"github.com/spf13/cast"
	"sync"
	"time"
)

var _ Cache = (*_redis)(nil)

type _redis struct {
	sync.Mutex
	ctx    context.Context
	client *redis.Client
}

func newRedis(ctx context.Context, client *redis.Client) *_redis {
	r := &_redis{ctx: ctx, client: client}

	go func() {
		<-r.ctx.Done()
		if client != nil {
			r.Close()
		}
	}()

	return r
}

func (r *_redis) GetDel(ctx context.Context, key string) ([]byte, error) {
	s, err := r.client.GetDel(ctx, key).Result()
	if err != nil {
		if errors.Is(err, redis.Nil) {
			return nil, ErrorKeyNotFound
		}

		return nil, err
	}

	return tool.StringToBytes(s), nil
}

func (r *_redis) GetDelScan(ctx context.Context, key string) Scanner {
	bs, err := r.GetDel(ctx, key)
	return newScan(bs, err)
}

func (r *_redis) Get(ctx context.Context, key string) ([]byte, error) {
	result, err := r.client.Get(ctx, key).Result()
	if err != nil {
		if errors.Is(err, redis.Nil) {
			return nil, ErrorKeyNotFound
		}

		return nil, err
	}

	return tool.StringToBytes(result), nil
}

func (r *_redis) Gets(ctx context.Context, keys ...string) ([][]byte, error) {
	result, err := r.client.MGet(ctx, keys...).Result()
	if err != nil {
		if errors.Is(err, redis.Nil) {
			return nil, ErrorKeyNotFound
		}

		return nil, err
	}

	return tool.Map(
		result,
		func(item any, index int) []byte {
			return tool.StringToBytes(cast.ToString(item))
		},
	), nil
}

func (r *_redis) GetScan(ctx context.Context, key string) Scanner {
	return newScan(r.Get(ctx, key))
}

func (r *_redis) GetEx(ctx context.Context, key string, duration time.Duration) ([]byte, error) {
	result, err := r.client.GetEx(ctx, key, duration).Result()
	if err != nil {
		if errors.Is(err, redis.Nil) {
			return nil, ErrorKeyNotFound
		}

		return nil, err
	}

	return tool.StringToBytes(result), nil
}

func (r *_redis) GetExScan(ctx context.Context, key string, duration time.Duration) Scanner {
	return newScan(r.GetEx(ctx, key, duration))
}

func (r *_redis) Set(ctx context.Context, key string, value any) error {
	bs, err := handleValue(value)
	if err != nil {
		return err
	}

	_, err = r.client.Set(ctx, key, bs, redis.KeepTTL).Result()
	return err
}

func (r *_redis) Sets(ctx context.Context, values map[string]any) error {
	vm := make(map[string]any)
	for k, v := range values {
		bs, err := handleValue(v)
		if err != nil {
			return err
		}

		vm[k] = bs
	}

	return r.client.MSet(ctx, vm).Err()
}

func (r *_redis) SetEx(ctx context.Context, key string, value any, duration time.Duration) error {
	bs, err := handleValue(value)
	if err != nil {
		return err
	}

	_, err = r.client.SetEX(ctx, key, bs, duration).Result()

	return err
}

func (r *_redis) Del(ctx context.Context, keys ...string) error {
	return r.client.Del(ctx, keys...).Err()
}

func (r *_redis) Close() {
	r.Lock()
	defer r.Unlock()
	_ = r.client.Close()
	r.client = nil
}
